-----------------------------------------------------------------------------
--
--  DATA_DISPATCHER_2.VHD
--
--	Component for distributing data to and from all connected Devices.
--  It Holds a Main Statemachine, several Muxers and allows to Retrieve Information
--  from connected devices. It also allows to reset devices selective (not tested)
--
--	Author 	: Rene Bllhoff
--  Created : around Mar. 2006 (Template Created)
--
--	(c) 2005-2007 by Rene Bllhoff - reneboellhoff (at) gmx (dot) de - 
-----------------------------------------------------------------------------
--
--  This source file may be used and distributed without
--  restriction provided that this copyright statement is not
--  removed from the file and that any derivative work contains
--  the original copyright notice and the associated disclaimer.
--
--  This program is free software; you can redistribute it and/or
--  modify it under the terms of the GNU General Public License
--  as published by the Free Software Foundation; either version 2
--  of the License, or (at your option) any later version.
--										
--  This program is distributed in the hope that it will be useful,
--  but WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the GNU
--  General Public License for more details.
--
--  You should have received a copy of the GNU General Public
--  License along with this source; if not, download it	from
--  http://www.gnu.org/licenses/gpl.html	
--
-----------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;


entity data_dispatcher_2 is

    Port ( clk50m 	: in	 	std_logic;
           res 			: in	 	std_logic;

					 -- Data Dispatcher Interface
					 dp_din   : in    std_logic_vector (7 downto 0);
					 dp_dout  :	out   std_logic_vector (7 downto 0);
					 dp_trg		: in    std_logic;
					 dp_sel		: in    std_logic;
 
					 -- common signals
					 sysreset :	out   std_logic;
					 dev_dout	: out std_logic_vector ( 7 downto 0);
 
					 -- Data Dispatcher Port 1
					 dev1_sel	 : out std_logic;
					 dev1_rdy	 : in  std_logic;
					 dev1_trg	 : out std_logic;
					 dev1_res  : out std_logic;
					 dev1_din	 : in  std_logic_vector ( 7 downto 0);
					 dev1_inf  : in  std_logic_vector (15 downto 0);

					 -- Data Dispatcher Port 2
					 dev2_sel	 : out std_logic;
					 dev2_rdy	 : in  std_logic;
					 dev2_trg	 : out std_logic;
					 dev2_res  : out std_logic;
					 dev2_din	 : in  std_logic_vector ( 7 downto 0);
					 dev2_inf  : in  std_logic_vector (15 downto 0)

				 );

end data_dispatcher_2;

architecture Behavioral of data_dispatcher_2 is


---------------------------------------------
--
-- 	Signals
--
---------------------------------------------	

	-- data out from dispatcher (version,info) (when data_route = 0)
	signal ddp_dataout   		: std_logic_vector (7 downto 0);	

	-- trigger signal and edge detection for dp_Trg
	signal fsm_trigger 		: std_logic;
	signal dp_trg_z1			: std_logic;
	
	-- state of dispatcher
	signal cmd_state  		: std_logic_vector (3 downto 0);
	-- routing of data in from devices (or dispatcher itsef)
	signal data_route     : std_logic_vector (3 downto 0);
	-- state to allow use dispatcher after device has quitted end of operation with rdy
	signal ctrl       		: std_logic_vector (3 downto 0);
	-- ready signal from current selected device
	signal fsm_is_ready   : std_logic;

	-- port info route signals
	signal info_route	 		: std_logic_vector ( 3 downto 0);
	signal info_data			: std_logic_vector (15 downto 0);

	-- reset signals (not tested)
	signal reset_mask     : std_logic_vector ( 7 downto 0);
	signal reset_system   : std_logic;

	-- version "string" consists of 1 byte X"22" means Dispatcher V2 / 2 Ports 
	constant dispatcher_info : std_logic_vector ( 7 downto 0) := X"22";	 -- high nibble = version ; low nibble = num ports

begin

  process (	clk50m, res)
	begin
    if res = '1' then
			-- Global Asynchronous reset			
			cmd_state				<= (others => '0');
			data_route			<= (others => '0');
			info_route			<= (others => '0');
			info_data				<= (others => '0');
			ctrl        		<= (others => '0');
			ddp_dataout			<= (others => '0');
			reset_mask			<= (others => '0');
			reset_system		<= '0';

			dev1_sel				<= '0';
			dev1_trg  			<= '0';

			dev2_sel				<= '0';
			dev2_trg  			<= '0';


			fsm_is_ready  	<= '1';

		elsif clk50m'event and clk50m = '1' then
	
			-- edge detection for main trigger of data dispatcher state machine
		  dp_trg_z1 <= dp_trg;
			fsm_trigger <= dp_trg and not dp_trg_z1;


			if reset_system = '1' then
			  -- synchronous reset (must not be implemented for every device, only if useful, an asynchronous would fit it :-))
				reset_system			<= '0';	

				cmd_state				<= (others => '0');
				data_route			<= (others => '0');
				info_route			<= (others => '0');
				info_data				<= (others => '0');
				ctrl        		<= (others => '0');
				ddp_dataout			<= (others => '0');
				fsm_is_ready  	<= '1';

			  -- More Ports may be added here
				dev1_sel				<= '0';
				dev1_trg  			<= '0';

				dev2_sel				<= '0';
				dev2_trg  			<= '0';



			end if;

			reset_mask			<= X"00";
			reset_system		<= '0';

			-- only if dispatcher is selected
			if dp_sel = '0' then
				-- ... and the whole thing is triggered
				if fsm_trigger = '1' then
				  -- ... do the statemachine
				  case cmd_state is
					  -- State 0 : The First Byte after CS has gone low decides to which Port Incoming Data is routed (resp. 
						--					 read From and send to Dispatcher Data Out) or gives access to the internal registers 
						--					 of the Data Dispatcher	
						when X"0" => case dp_din (7 downto 0) is
												 	 -- Port Selection
												   when X"F0"   	=> cmd_state <= X"1"; 	data_route <= X"1"; 
												   when X"F1"   	=> cmd_state <= X"1"; 	data_route <= X"2"; 
											 		 -- More Ports may be added here
														 								 
												 	 -- Port Info
													 when X"E0"   	=> cmd_state <= X"2"; 	data_route <= X"0"; 															

												   -- Dispatcher Info 
													 when X"ED"   	=> cmd_state <= X"5"; 	data_route <= X"0"; 															

												   -- Reset Selective (not tested)
													 when X"EE"   	=> cmd_state <= X"6"; 	data_route <= X"0"; 															

												   -- Reset All (not tested)
													 when X"EF"   	=> cmd_state <= X"7"; 	data_route <= X"0"; 															

													 when others  	=> cmd_state <= X"0"; 	data_route <= X"0";
												 end case; 	

						-- State 1 : simply do nothing until fsm_is_ready is set (and ctrl goes in idle)
						--					 all incoming data is routed to the previous selected port
						when X"1" => NULL;						 

						-- Port Info
					
						-- State 2 : Next Byte in stream selects Port to read Information-Word from
						when X"2"	=> info_route   		<= dp_din (3 downto 0); 			cmd_state <= X"3"; 
						-- Stste 3 : Prepares High-Byte of Information-Word to transmit
						when X"3" => ddp_dataout			<= info_data (15 downto 8); 	cmd_state <= X"4"; 
						-- Stste 4 : Prepares Low-Byte of Information-Word to transmit
						when X"4" => ddp_dataout			<= info_data ( 7 downto 0); 	cmd_state <= X"0"; 


						-- Stste 5 : Read Constant dispatcher_info ( X"22" means Version 2; 2 Ports)
						when X"5" => ddp_dataout			<= dispatcher_info;						cmd_state <= X"0";

						-- Stste 6 : Reset Selective
						when X"6" => reset_mask				<= dp_din;										cmd_state <= X"0";

						-- Stste 7 : Reset All
						when X"7"	=> reset_system			<= '1';	reset_mask <= X"ff"; 	cmd_state <= X"0";

						when others => cmd_state 			<= X"0";
					end case;
	      end if;

				--
				-- Statemachine to wait for fsm_is_ready to allow Port Reselection on Opeartion End (not tested)
				--
	      case ctrl is
	        when X"0" => if data_route /= X"0" then		 -- wait for fsm_is_ready = 0 when port is selected 
	                       if fsm_is_ready = '0' then
	                         ctrl <= X"1";  
	                       end if;
	                     end if;
				
	        when X"1" => if data_route /= X"0" then		 -- wait for fsm_is_ready = 1  -> reset statemachine and wait for Port Selection
	                       if fsm_is_ready = '1' then
	                         ctrl <= X"0";  
				                	 dp_dout <= (others => '0');
				                   cmd_state <= X"0"; 
				               	 data_route 		<= X"0";
	                      end if;
	                    end if;
						
	          when others => NULL;
	      end case;  
				 
				--
				-- Get Info from Port
				--				
				case info_route is
				  when X"0"   => info_data <= dev1_inf;
				  when X"1"   => info_data <= dev2_inf;
					-- More Ports may be added here
					when others => NULL;
				end case;				   
       

			else
			  -- if dispatcher is deselected ->
				-- immediately deselect and reset all routes and states
			  data_route 	<= X"0"; 
				cmd_state 	<= X"0"; 
				ctrl 				<= X"0"; 
				ddp_dataout	<= (others => '0');
			end if;
		end if;


		-- Enable Port depening on data_route
		if data_route = X"1" then dev1_sel <= '1'; else dev1_sel <= '0'; end if;
		if data_route = X"2" then dev2_sel <= '1'; else dev2_sel <= '0'; end if;
		-- More Ports may be added here


		--
    -- Main Muxer (routes Data Port, Ready and Trigger - Signals)
		--
		case data_route is
			when X"1" 	 => dev1_trg		 	<= fsm_trigger; 
			                fsm_is_ready 	<= dev1_rdy; 
			                dp_dout  			<= dev1_din;

			when X"2" 	 => dev2_trg 		 	<= fsm_trigger; 
			                fsm_is_ready 	<= dev2_rdy; 
			                dp_dout  			<= dev2_din;

			-- More Ports may be added here
			            
			-- Default Port (0) selected when Information from Dispatcher is Read	(via ddp_dataout)
			when others  => fsm_is_ready 	<= '1'; 					     
 			                dp_dout		  	<= ddp_dataout;

		end case;	

	end process;

	-- simply output signals from internal registers

	sysreset    <= reset_system;
	dev1_res		<= reset_mask (0);
	dev2_res		<= reset_mask (1);
	dev_dout 		<= dp_din;

end Behavioral;
