-----------------------------------------------------------------------------
--
--	UART_CLOCK.VHDL
--
--  Generates Baud Clocks (1x, 16x) from System Clock
--  The main Divider can be set to a individual Value or a Value which results 
--  in a "Base Baud Rate" (aprox. 1.851 MHz) and then use fixed dividers to
--  simply select std-baud rates (as 38400 or 57600, downto 1200) via lowest
--  nibble in the 16bit-wide baudrate-signal 
--
--
--	Author 	: Rene Bllhoff
--  Created : Feb. 2007
--
--	(c) 2005-2007 by Rene Bllhoff - reneboellhoff (at) gmx (dot) de - 
-----------------------------------------------------------------------------
--
--  This source file may be used and distributed without
--  restriction provided that this copyright statement is not
--  removed from the file and that any derivative work contains
--  the original copyright notice and the associated disclaimer.
--
--  This program is free software; you can redistribute it and/or
--  modify it under the terms of the GNU General Public License
--  as published by the Free Software Foundation; either version 2
--  of the License, or (at your option) any later version.
--										
--  This program is distributed in the hope that it will be useful,
--  but WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the GNU
--  General Public License for more details.
--
--  You should have received a copy of the GNU General Public
--  License along with this source; if not, download it	from
--  http://www.gnu.org/licenses/gpl.html	
--
-----------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;


entity uart_clock is
    Port ( clk 					: in 	std_logic;
           res 					: in 	std_logic;
           baudrate 		: in 	std_logic_vector(15 downto 0);
           baud_clk_o 	: out std_logic;
           baud_clk16_o : out std_logic);
end uart_clock;

architecture Behavioral of uart_clock is

	-- counter values and main clocks
  signal base_baud_counter 	: std_logic_vector (14 downto 0);	-- for individual baudrate
  signal base_baud_clk  		: std_logic;											-- 
  signal baud_clk_16    		: std_logic;  	
  signal baud_clk       		: std_logic;  	
  signal baud_count_max 		: std_logic_vector ( 7 downto 0); -- max 2048 => ca 452 baud min.
  signal baud_count     		: std_logic_vector ( 7 downto 0); -- baudraten vorteiler (aus 1.851 MHz)
  signal baud_count_tx			: std_logic_vector ( 3 downto 0);

	-- signals for edge detection
	signal baud_clk_16_z1			: std_logic;
	signal base_baud_clk_z1		: std_logic;
	
begin

  baud_clk_o   	<= baud_clk;
	baud_clk16_o 	<= baud_clk_16;

	-- "baudraten" basis generator (57600 * 16 * 2) ca. 1.851 MHz
	process (clk, res)
	begin
	  --divier_value_std_baud <= conv_std_logic_vector (base_prescaler, 8);

	  if res = '1' then
			base_baud_counter <= (others => '0');
	    base_baud_clk 		<= '0';      

	  elsif clk'event and clk = '1' then
			base_baud_counter <= base_baud_counter + '1';
			-- uses divider to get to 1.851mhz ?
			if baudrate (15) = '0' then
		    if base_baud_counter = baudrate (11 downto 4) then
		      base_baud_clk <= '1';   
					base_baud_counter <= (others => '0');   
		    else
		      base_baud_clk <= '0';      
		    end if;
			else
			  -- or individual divider
		    if base_baud_counter = baudrate (14 downto 0) then
		      base_baud_clk <= '1';   
					base_baud_counter <= (others => '0');   
		    else
		      base_baud_clk <= '0';      
		    end if;
			end if;
	  end if;
	end process;

	-- baudrate generator RX (16x baud clock)
	process (clk, res)
	begin
	  if res = '1' then
	    baud_count  			<= (others => '0');
	    baud_clk_16 			<= '0';   
			base_baud_clk_z1 	<= '0';   

	  elsif clk'event and clk = '1' then

			-- edge detection																					
 		  base_baud_clk_z1 <= base_baud_clk;
	    if base_baud_clk_z1 = '0' and base_baud_clk = '1' then
				-- if we use 1.851mhz, do fixed divider
				if baudrate (15) = '0' then
				  -- do the dixed divider
					baud_count <= baud_count + '1';
			    if baud_count = baud_count_max then
			      baud_count  <= (others => '0');      
			      baud_clk_16 <= '1';
			    else
			      baud_clk_16 <= '0';      
			    end if;
				else
				  -- otherwise use clock from individual divier as baud_clk_16
					baud_clk_16 <= '1';
				end if;
			else
			  -- generate baud_clk_16 with width of system clock
				baud_clk_16 <= '0';      
			end if;

	  end if;
	end process;

	-- baudrate generator for TX
	process (clk, res)
	begin
	  if res = '1' then
	    baud_count_tx  	<= (others => '0');
	    baud_clk 				<= '0'; 
			baud_clk_16_z1  <= '0';

	  elsif clk'event and clk = '1' then

			-- edge detection																					
			baud_clk_16_z1  <= baud_clk_16; 
		  if baud_clk_16_z1 = '0' and baud_clk_16 = '1' then
				-- simply divide by 16 ...
		    baud_count_tx <= baud_count_tx + '1';
		    if baud_count_tx = X"0" then
		      baud_clk <= '1';
		    else
		      baud_clk <= '0';      
		    end if;
			else
			  -- generate baud_clk_16 with width of system clock
	      baud_clk <= '0';      
			end if;
	  end if;
	end process;


	-- value for fixed divider (when using 1.851MHz)
	process (clk, res)
	begin  
	  if res = '1' then
	    baud_count_max <= X"02"; -- by default use 38400 baud

	  elsif clk'event and clk = '1' then
	    if baud_clk = '1' then
	      case baudrate (3 downto 0) is
	        when X"0" => baud_count_max <= X"5f"; -- 1200 Baud
	        when X"1" => baud_count_max <= X"2f"; -- 2400 Baud
	        when X"2" => baud_count_max <= X"17"; -- 4800 Baud
	        when X"3" => baud_count_max <= X"0b"; -- 9600 Baud
	        when X"4" => baud_count_max <= X"05"; -- 19200 Baud
	        when X"5" => baud_count_max <= X"02"; -- 38400 Baud
	        when X"6" => baud_count_max <= X"01"; -- 57600 Baud
	        when others => NULL;
	      end case;
	    end if;
	  end if;
	end process;


end Behavioral;
