-----------------------------------------------------------------------------
--
--  I2S_SHIFTER_2x_SLAVE.VHDL
--
--	Implements an 2x I2S Interface. The I2S Interface behaves as a slave.
--  Clocking has to be done externally. 
--
--	Author 	: Rene Bllhoff
--  Created : around Jan. 2007
--
--	(c) 2005-2007 by Rene Bllhoff - reneboellhoff (at) gmx (dot) de - 
-----------------------------------------------------------------------------
--
--  This source file may be used and distributed without
--  restriction provided that this copyright statement is not
--  removed from the file and that any derivative work contains
--  the original copyright notice and the associated disclaimer.
--
--  This program is free software; you can redistribute it and/or
--  modify it under the terms of the GNU General Public License
--  as published by the Free Software Foundation; either version 2
--  of the License, or (at your option) any later version.
--										
--  This program is distributed in the hope that it will be useful,
--  but WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the GNU
--  General Public License for more details.
--
--  You should have received a copy of the GNU General Public
--  License along with this source; if not, download it	from
--  http://www.gnu.org/licenses/gpl.html	
--
-----------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity i2s_shifter_2x_slave is
    Port ( clk50m 			: in std_logic;
					 res	 				: in std_logic;

					 -- I2S I/O Signals
					 -- Common Clocks
					 bclk 				: in std_logic;
           lrclk 				: in std_logic;
					 -- Interface 1
           sdin1				: in std_logic;
					 sdout1				: out std_logic;
					 -- Interface 2
           sdin2				: in std_logic;
					 sdout2				: out std_logic;

					 -- Global Mute (affects only output)
					 mute					: in std_logic;
	
					 -- New Sample
           new_sample  	: out std_logic;

					 -- 4 parallel inputs (to I2S Output)
           sample_1_in 	: in  signed(23 downto 0);
           sample_2_in 	: in  signed(23 downto 0);
           sample_3_in 	: in  signed(23 downto 0);
           sample_4_in 	: in  signed(23 downto 0);
					 -- 4 parallel outputs (frmo I2S Input)
           sample_1_out : out signed(23 downto 0);
           sample_2_out : out signed(23 downto 0);
           sample_3_out : out signed(23 downto 0);
           sample_4_out : out signed(23 downto 0));
end i2s_shifter_2x_slave;





architecture Behavioral of i2s_shifter_2x_slave is

	-- define states as type
	type i2s_shift_states is (stInit, 				-- From Reset
														stWaitLrEdge, 	-- Waits for L/R-Change and copies Data from/to In-/Outputs on L-R-Change (falling edge)
														stNewSmp, 			-- 
														stShiftOut, 		-- shifts one bit out of register
														stShiftIn, 			-- shifts one bit into register
														stIncBitCnt 
														);

	-- state as a signal
	signal i2s_state				  : i2s_shift_states;

	-- input and output buffers (24 bit)
	signal in_1_buffer  			: std_logic_vector (23 downto 0);
	signal in_2_buffer  			: std_logic_vector (23 downto 0);
	signal in_3_buffer  			: std_logic_vector (23 downto 0);
	signal in_4_buffer  			: std_logic_vector (23 downto 0);
	signal out_1_buffer 			: std_logic_vector (23 downto 0);
	signal out_2_buffer 			: std_logic_vector (23 downto 0);
	signal out_3_buffer 			: std_logic_vector (23 downto 0);
	signal out_4_buffer 			: std_logic_vector (23 downto 0);

	-- internal channel signal to route ro register 1&3 or 2&4 
	signal channel						: std_logic;

	-- clocking in of l/r-clock and bit-clock
	signal lrclk_z1, lrclk_z2, lrclk_z3 : std_logic;
	signal bclk_z1, bclk_z2, bclk_z3 		: std_logic;

	-- temporarily mute (is only updated on new sample)
	signal temp_nmute					: std_logic;

	-- constant defining max. word-length
	constant numbits 				  : integer := 23;

begin

	process (clk50m, res) 
	  variable bitcnt : integer range 0 to 23;
	begin
	  if res = '1' then
			-- reset all buffers
			in_1_buffer  	<= (others => '0');
			in_2_buffer  	<= (others => '0');
			in_3_buffer  	<= (others => '0');
			in_4_buffer  	<= (others => '0');
			out_1_buffer 	<= (others => '0');
			out_2_buffer 	<= (others => '0');
			out_3_buffer 	<= (others => '0');
			out_4_buffer 	<= (others => '0');
			-- reset clocking in of lrclk and bclk
			bclk_z1 	 		<= '0';
			bclk_z2 	 		<= '0';
			bclk_z3 	 		<= '0';
			lrclk_z1 	 		<= '0';
			lrclk_z2 	 		<= '0';
			lrclk_z3 	 		<= '0';
			-- set outputs to 0 (i2s interface)
			sdout1 		 		<= '0';
			sdout2 		 		<= '0';
			-- set outputs to 0 (parallel interface)
			sample_1_out	<= (others => '0');
			sample_2_out	<= (others => '0');	
			sample_3_out	<= (others => '0');
			sample_4_out	<= (others => '0');	
			-- reset internal signals (and variables)
			bitcnt 		 		:= 0;
			i2s_state  		<= stInit;
			channel 			<= '0';
			new_sample		<= '0';
			temp_nmute		<= '1';

		elsif clk50m'event and clk50m = '1' then

			-- do clocking in of lrclk and bclk
			lrclk_z1 <= lrclk;
			lrclk_z2 <= lrclk_z1;
			lrclk_z3 <= lrclk_z2;

			bclk_z1  <= bclk;
			bclk_z2  <= bclk_z1;
			bclk_z3  <= bclk_z2;

			case i2s_state is
			  when stInit 				=> 	sdout1 <= '0';
																sdout2 <= '0';
																bitcnt := 0;
											 					i2s_State <= stWaitLREdge;
																new_sample <= '0';

			  when stWaitLREdge  => 	-- on falling edge of lrclk (start of frame) ...
																if lrclk_z3 = '1' and lrclk_z2 = '1' and  lrclk_z1 = '0' then

																	-- take over input and output buffers
																	sample_1_out <= signed (in_1_buffer);
																	sample_2_out <= signed (in_2_buffer);
																	sample_3_out <= signed (in_3_buffer);
																	sample_4_out <= signed (in_4_buffer);

																	out_1_buffer <= std_logic_vector (sample_1_in);
																	out_2_buffer <= std_logic_vector (sample_2_in);			
																	out_3_buffer <= std_logic_vector (sample_3_in);
																	out_4_buffer <= std_logic_vector (sample_4_in);			

																  i2s_State <= stNewSmp;	-- here we go to a between-state where we set our new-sample signal 
																	-- set outputs to 0
																	sdout1 <= '0';
																	sdout2 <= '0';

																	-- set internal signals
																	channel <= '0';
																	temp_nmute <= not mute;
																	new_sample <= '0';

																-- on rising edge of lrclk (second half of frame)...
																elsif lrclk_z3 = '0' and  lrclk_z2 = '0' and  lrclk_z1 = '1' then
																  i2s_State <= stShiftOut;
																	-- set outputs to 0
																	sdout1 <= '0';
																	sdout2 <= '0';

																	-- set internal signals
																	channel <= '1';
																	temp_nmute <= not mute;
																	new_sample <= '0';
																end if;
																new_sample <= '0';

				when stNewSmp				=>	new_sample <= '1';        -- rise new_sample for 1 clock
																i2s_State <= stShiftOut;
											 					
			  when stShiftOut 		=>  -- on  falling edge of bclk do shifting out
																if bclk_z3 = '1' and bclk_z2 = '1' and bclk_z1 = '0'then
																  -- depending on channel ...
																	if channel = '0' then
																		-- ... shift out channels 1&3 on i2s lines
																	  sdout1 <= out_1_buffer (numbits - bitcnt) and temp_nmute;
																	  sdout2 <= out_3_buffer (numbits - bitcnt) and temp_nmute;
																	else
																		-- ... or shift out channels 1&3 on i2s lines
																	  sdout1 <= out_2_buffer (numbits - bitcnt) and temp_nmute;
																	  sdout2 <= out_4_buffer (numbits - bitcnt) and temp_nmute;
																	end if;  
																	-- go shifting in
																  i2s_State <=  stShiftIn;
																else 
																  i2s_State <=  stShiftOut;
																end if;
  															new_sample <= '0'; -- reset new_sample

			  when stShiftIn 			=>  -- on  rising edge of bclk do shifting in
																if bclk_z3 = '0' and bclk_z2 = '0' and bclk_z1 = '1' then
																  -- depending on channel ...
																	if channel = '0' then
																	  -- ... shift in channels 1&3 from i2s lines
																		in_1_buffer (numbits - bitcnt) <= sdin1;
																		in_3_buffer (numbits - bitcnt) <= sdin2;
																	else
																		-- ... or shift in channels 1&3 from i2s lines
																		in_2_buffer (numbits - bitcnt) <= sdin1;
																		in_4_buffer (numbits - bitcnt) <= sdin2;
																	end if;
																	-- increment bit counter
																	i2s_State <=  stIncBitCnt;
																else
																  i2s_State <=  stShiftIn;
																end if;
																new_sample <= '0';

			  when  stIncBitCnt =>  	-- shifted all 24 bits in ?
																if bitcnt = 23 then
																	bitcnt := 0;

																	-- go waiting for another l/r-edge
											  					i2s_State <= stWaitLREdge;
																	
																	if channel = '1' then
																	  new_sample <= '0';
																	else
																	  new_sample <= '0';
																	end if;																	  

																else
																	-- increment bit counter and continue with shifting out
																	bitcnt := bitcnt + 1; 
											  					i2s_State <= stShiftOut;	
																	new_sample <= '0';
																end if;

  			when others					=> 	NULL;
			end case;
		end if;
	end process;

end Behavioral;