-----------------------------------------------------------------------------
--
--	DEVICE_STDIO_V2.VHDL
--
--	Implements a Top-Level-Entity to control the simple Hardware-Portion
--  (LEDs, Switches, Buttons and 7 Sgement Display) of the Spartan3-Development 
--  board. It allows to set Segments of 7Segment Display individually as well as
--  Displaying 16-Bit Hexadecimal. It also holds a possibilty
--  of using all components for Debugging purposes without reconnecting the
--  complete design.
--
--	Author 	: Rene Bllhoff
--  Created : around May 2006
--
--	(c) 2005-2007 by Rene Bllhoff - reneboellhoff (at) gmx (dot) de - 
-----------------------------------------------------------------------------
--
--  This source file may be used and distributed without
--  restriction provided that this copyright statement is not
--  removed from the file and that any derivative work contains
--  the original copyright notice and the associated disclaimer.
--
--  This program is free software; you can redistribute it and/or
--  modify it under the terms of the GNU General Public License
--  as published by the Free Software Foundation; either version 2
--  of the License, or (at your option) any later version.
--										
--  This program is distributed in the hope that it will be useful,
--  but WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the GNU
--  General Public License for more details.
--
--  You should have received a copy of the GNU General Public
--  License along with this source; if not, download it	from
--  http://www.gnu.org/licenses/gpl.html	
--
-----------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

library UNISIM;
use UNISIM.VComponents.all;

entity device_stdio_v2 is
    port (
        clk50m  		: in  std_logic;
        res     		: in  std_logic;  
				
				-- Device Port Signals
				fsm_select  : in  std_logic;
				fsm_trigger : in  std_logic;
				fsm_ready		: out std_logic;
				fsm_reset		: in  std_logic;
				fsm_din  		: in  std_logic_vector ( 7 downto 0);
				fsm_dout 		: out std_logic_vector ( 7 downto 0);
				fsm_inf	  	: out std_logic_vector (15 downto 0);

				-- Grabbing StdIO-Hardware-Parts for Debugging Purposes
				grab_7seg		: in 	std_logic;		-- hard route 7 seg (16hex & 32seg bit) for debugging
				grab_dots 	: in 	std_logic;		-- hard route 4 dots for debugging
				grab_leds		: in 	std_logic;		-- hard route 8 leds for debugging

				-- Grabbing options
				use_hex_in  : in 	std_logic;		-- use 16 bit interface
				use_dots_in	: in 	std_logic;		-- use dots
				data16in		: in 	std_logic_vector (15 downto 0); -- 16 bit hex-value
				data32in		: in 	std_logic_vector (31 downto 0);	-- 32 bit segment
				dots4in     : in 	std_logic_vector ( 3 downto 0);	-- 4 dots
				leds8in     : in 	std_logic_vector ( 7 downto 0); -- 8 eds

				-- I/O Interfae to Spartan 3 Onboard Hardware
				swin				:	in	std_logic_vector ( 7 downto 0);
				pbtn				:	in	std_logic_vector ( 2 downto 0);	
				ledout			:	out std_logic_vector ( 7 downto 0);
				segout			:	out std_logic_vector ( 7 downto 0);
				digout			:	out std_logic_vector ( 3 downto 0)
		);
end device_stdio_v2;



architecture Behavioral of device_stdio_v2 is

---------------------------------------------
--
--  7x4 Segment Muxer Component
--
---------------------------------------------	

	COMPONENT seg7mux_v2
	PORT(
		clk   		: IN 	std_logic;
		clk1k 		: IN 	std_logic;
		res 			: IN 	std_logic;
		use_hex 	: IN 	std_logic;
		use_dots 	: IN 	std_logic;
		value16 	: IN 	std_logic_vector(15 downto 0);
		value32 	: IN 	std_logic_vector(31 downto 0);
		dots4 		: IN 	std_logic_vector(3 downto 0);          
		segout 		: OUT std_logic_vector(7 downto 0);
		digout 		: OUT std_logic_vector(3 downto 0)
		);
	END COMPONENT;

---------------------------------------------
--
-- Clock Generation for 7 Segment Component
--
---------------------------------------------	


	COMPONENT clockgen   
	PORT(
		clk50m  : IN  std_logic;
		res			: IN 	std_logic;
		clk1mo	: OUT std_logic;
		clk10ko	: OUT std_logic;
		clk1ko	: OUT std_logic;
		clk100o	: OUT std_logic;
		clk1o		: OUT std_logic
		);	
	END COMPONENT;


---------------------------------------------
--
--  Signals
--
---------------------------------------------	



	signal clock1k  				: std_logic;

	signal data16  					: std_logic_vector (15 downto 0);
	signal data32  					: std_logic_vector (31 downto 0);
	signal leds		 					: std_logic_vector ( 7 downto 0);	
	signal dots4s  					: std_logic_vector ( 3 downto 0);

	signal data16_t  				: std_logic_vector (15 downto 0);
	signal data32_t 				: std_logic_vector (31 downto 0);
	signal dots_t	 					: std_logic_vector ( 3 downto 0);	

	signal use_hex  				: std_logic;
	signal use_hex_t				: std_logic;
	signal use_dots 				: std_logic;
	signal use_dots_t				: std_logic;


	signal pbtn_z1, pbtn_z2 : std_logic_vector ( 2 downto 0);
	signal swin_z1, swin_z2 : std_logic_vector ( 7 downto 0);

	signal device_state			: std_logic_vector ( 3 downto 0);

begin

---------------------------------------------
--
-- Instanciation of Components
--
---------------------------------------------	

	-- Clock Generation for 7 Segment Muxer
	Inst_clockgen : clockgen port map (
		clk50m 	=> clk50m,
		res			=> res,
		clk1ko	=> clock1k
	);

	-- Main Component for 4 7 Segment Multiplexer
	Inst_seg7muxer_v2 : seg7mux_v2 port map (
		clk   		=> clk50m, 
		clk1k 		=> clock1k, 
		res 			=> res, 
		use_hex		=> use_hex_t,
		use_dots	=> use_dots_t,
		value16 	=> data16_t,
		value32 	=> data32_t,
		dots4 		=> dots_t,
		segout 		=> segout, 
		digout 		=> digout
	);


---------------------------------------------
--
--  Statemachine for Accessing StdIO Hardware
--
---------------------------------------------	

  

  process (clk50m, res)
	begin
    if res = '1' then
			
			fsm_dout 	<= (others => '0');								
			device_state <= x"0";
			fsm_ready 		<= '1';


			data16 				<= (others => '0');
			data16_t			<= (others => '0');
			data32 				<= (others => '0');
			data32_t			<= (others => '0');
			leds	 				<= (others => '0');
			dots4s 				<= (others => '0');
			dots_t				<= (others => '0');

			use_hex				<= '1';
			use_hex_t			<= '1';
			use_dots			<= '1';
			use_dots_t		<= '1';

			pbtn_z1				<= (others => '0');
			pbtn_z2				<= (others => '0');

			swin_z1				<= (others => '0');
			swin_z2				<= (others => '0');


		elsif clk50m'event and clk50m = '1' then

			pbtn_z1 <= pbtn;
		  pbtn_z2 <= pbtn_z1;

			swin_z1 <= swin;
			swin_z2 <= swin_z1;


		  if fsm_reset = '1' then
				fsm_dout 	<= (others => '0');								
				device_state <= x"0";
				fsm_ready 		<= '1';


				data16 				<= (others => '0');
				data32 				<= (others => '0');
				leds	 				<= (others => '0');
				dots4s 				<= (others => '0');

			end if;

      if fsm_select = '1' then
         fsm_ready  <= '0';

				 if fsm_trigger = '1' then

					 case device_state	is
						 when X"0" => 	case fsm_din (7 downto 0) is
							 								when X"00" => device_state <= X"1"; fsm_ready <= '0';
															when X"01" => device_state <= X"2"; fsm_ready <= '0';
															when X"02" => device_state <= X"3"; fsm_ready <= '0';
															when X"03" => device_state <= X"4"; fsm_ready <= '0';
															when X"04" => device_state <= X"5"; fsm_ready <= '0'; fsm_dout <= swin;
															when X"05" => device_state <= X"6"; fsm_ready <= '0'; fsm_dout <= X"0" & '0' & pbtn;

							 								when X"06" => device_state <= X"7"; fsm_ready <= '0';
															when X"07" => device_state <= X"8"; fsm_ready <= '0';
															when X"08" => device_state <= X"9"; fsm_ready <= '0';
															when X"09" => device_state <= X"a"; fsm_ready <= '0';
															when X"0a" => device_state <= X"b"; fsm_ready <= '0'; 

															when others=> device_state <= X"0"; fsm_ready <= '1'; fsm_dout <= (others => '0');
							 							end case;
						 when X"1" => data16 (15 downto 8) <= fsm_din; 	fsm_ready <= '1'; device_state <= X"0";
						 when X"2" => data16 ( 7 downto 0) <= fsm_din; 	fsm_ready <= '1'; device_state <= X"0";
						 when X"3" => dots4s  <= fsm_din (3 downto 0); 	fsm_ready <= '1'; device_state <= X"0";
						 when X"4" => leds                 <= fsm_din; 	fsm_ready <= '1'; device_state <= X"0";
						 when X"5" =>                                  	fsm_ready <= '1'; device_state <= X"0";
						 when X"6" =>                                  	fsm_ready <= '1'; device_state <= X"0";

						 when X"7" => use_hex  <= fsm_din (0);
						 							use_dots <= fsm_din (1);					fsm_ready <= '1'; device_state <= X"0";
						 when X"8" => data32 (31 downto 24) <= fsm_din; fsm_ready <= '1'; device_state <= X"0";
						 when X"9" => data32 (23 downto 16) <= fsm_din; fsm_ready <= '1'; device_state <= X"0";
						 when X"a" => data32 (15 downto  8) <= fsm_din; fsm_ready <= '1'; device_state <= X"0";
						 when X"b" => data32 ( 7 downto  0) <= fsm_din; fsm_ready <= '1'; device_state <= X"0";


						 when others => fsm_dout <= (others => '0');		fsm_ready <= '1';	device_state <= X"0";
					 end case;

         end if;  

      else
        fsm_ready <= '1';
			  device_state <= X"0";
			  fsm_dout <= (others => '0');
			end if;

			--
			-- do grabbing
			--

			if grab_7seg = '1' then
			  data16_t 		<= data16in;
				data32_t 		<= data32in;
				use_hex_t		<= use_hex_in;
				use_dots_t	<= use_dots_in;
			else
			  data16_t 		<= data16;
				data32_t 		<= data32;
				use_hex_t		<= use_hex;
				use_dots_t	<= use_dots;
			end if;

			if grab_dots = '1' then
				dots_t			<= dots4in;
			else
				dots_t			<= dots4s;
			end if;

			if grab_leds = '1' then
			  ledout <= leds8in;
			else
			  ledout <= leds;
			end if;
   
    end if; 

		-- Device Information "String" 1120
		--
		-- "1" = StdIO Hardware  "1" = StdIO Hardware (LEDs & Switches) "20" = Version 2.0
		-- 

		fsm_inf <= X"1120";

	end process;

end Behavioral;
